/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.tokenFifo;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.FleetGlobal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

public class TokenFifoShipElement extends Component {
	/**
	 * This is a basic builidng block for a Fifo. A string of two FifoElement-s
	 * forms a one-place fifo, three fifo elements form a two place fifo, ...
	 * 
	 * Interface: FleetGlobal.IN, FleetGlobal.ACK_IN on the data input side, 
	 * and FleetGlobal.OUT, FleetGlobal.ACK_OUT on the output side.
	 * 
	 * Token validity is not checked.
	 * 
	 * @author ib27688
	 */
 

	InputTerminal in;
	OutputTerminal ackIn;
	
	OutputTerminal tokenOut;
	InputTerminal ackTokenOut;
	
	OutputTerminal out;
	InputTerminal ackOut;
	
	InputTerminal tokenIn;
	OutputTerminal ackTokenIn;
	
	TokenFifoAction action;
	
	public TokenFifoShipElement(String name, Parameters params) {
		// no need for an variable name here

		super(name, null);

		// process the parameters
		Delay fwdDelayParam= null;
		Delay ackDelayParam= null;
		Delay tokenDelayParam= null;
		Delay ackTokenDelayParam= null;
		
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
				
				if (pName.equals(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY) 
						|| pName.equals(TokenFifoShipDefaults.FWD_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						fwdDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY) 
						|| pName.equals(TokenFifoShipDefaults.ACK_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						ackDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.TOKEN_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						tokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else if (pName.equals(TokenFifoShipDefaults.ACK_TOKEN_DELAY)) {
					try {
						int d = Integer.parseInt(sValue);
						ackTokenDelayParam= new Delay(d);
					}
					catch (Exception e) {
						director.error("Component: " + name + ", id= " + getID() 
								+ ". Bogus value used for " 
								+ TokenFifoShipDefaults.ACK_TOKEN_DELAY
								+ ": " + sValue
								+ ". Value is IGNORED");						
					}
				}
				else {
					director.error("Component: " + name + ", id= " + getID() 
							+ ": Attribute " + pName + " with value " + sValue
							+ " is ignored");					
				} // else
			} // for
		} // if params != null		
		
		buildElement(-1);
		
		// set delays if needed
		if (fwdDelayParam != null) {
			action.passDataFn.setDelay(fwdDelayParam);
		}
		if (ackDelayParam != null) {
			action.ackInFn.setDelay(ackDelayParam);
		}			
		if (tokenDelayParam != null) {
			action.tokenOutFn.setDelay(tokenDelayParam);
		}			
		if (ackTokenDelayParam != null) {
			action.ackTokenInFn.setDelay(ackTokenDelayParam);
		}			
	} // constructor
	
	
	public TokenFifoShipElement(int i) {
		// no need for an variable name here
		super("TokenFifoElement", null);
		
		buildElement(i);

	} // TokenFifoElement
	
	protected void buildElement(int i) {
		if (i >= 0) {
			action= new TokenFifoAction("TokenFifoAction_" + i);
		} 
		else {
			action= new TokenFifoAction("TokenFifoAction_");
		}
		addMember(action);
		
		in= new InputTerminal(FleetGlobal.IN);
		addMember(in);
		action.attach(in);
		
		ackIn= new OutputTerminal(FleetGlobal.ACK_IN);
		addMember(ackIn);
		action.attach(ackIn);
		
		tokenOut= new OutputTerminal(FleetGlobal.TOKEN_OUT);
		addMember(tokenOut);
		action.attach(tokenOut);
		
		ackTokenOut= new InputTerminal(FleetGlobal.ACK_TOKEN_OUT);
		addMember(ackTokenOut);
		action.attach(ackTokenOut);
		
		
		out= new OutputTerminal(FleetGlobal.OUT);
		addMember(out);
		action.attach(out);
		
		ackOut= new InputTerminal(FleetGlobal.ACK_OUT);
		addMember(ackOut);
		action.attach(ackOut);
		
		tokenIn= new InputTerminal(FleetGlobal.TOKEN_IN);
		addMember(tokenIn);
		action.attach(tokenIn);
		
		ackTokenIn= new OutputTerminal(FleetGlobal.ACK_TOKEN_IN);
		addMember(ackTokenIn);
		action.attach(ackTokenIn);

				
	}
	
	public Terminal getTerminal(String name) {
		if (name.equals(FleetGlobal.IN)) {
			return in;
		}
		else if (name.equals(FleetGlobal.ACK_IN)) {
			return ackIn;
		}
		else if (name.equals(FleetGlobal.OUT)) {
			return out;
		}
		else if (name.equals(FleetGlobal.ACK_OUT)) {
			return ackOut;
		}
		if (name.equals(FleetGlobal.TOKEN_IN)) {
			return tokenIn;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_IN)) {
			return ackTokenIn;
		}
		else if (name.equals(FleetGlobal.TOKEN_OUT)) {
			return tokenOut;
		}
		else if (name.equals(FleetGlobal.ACK_TOKEN_OUT)) {
			return ackTokenOut;
		}
		else
			return null;
	} // getTerminal

	
	class TokenFifoAction extends Action {
		
		int counter= 0;
		public PassData passDataFn;
		public ActionFired ackInFn;
		public ActionFired tokenOutFn;
		public ActionFired ackTokenInFn;
		
		public TokenFifoAction(String name) {
			super(name);
			
			passDataFn= new PassData();
			ackInFn= new ActionFired();
			tokenOutFn= new ActionFired();
			ackTokenInFn= new ActionFired();
			
			// get default delays for ack and for tokens
			Integer fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY);
			Delay ackInDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY_DEF;			
			ackInFn.setDelay(ackInDelay);	
			
			fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_TOKEN_DELAY);
			Delay tokenDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_TOKEN_DELAY_DEF;			
			tokenOutFn.setDelay(tokenDelay);	

			fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_TOKEN_DELAY);
			Delay ackTokenDelay= (fDelay != null)
				? new Delay(fDelay.intValue())
				: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_TOKEN_DELAY_DEF;			
			ackTokenInFn.setDelay(ackTokenDelay);	
			
			outputFunctions.put(FleetGlobal.OUT, passDataFn);
			outputFunctions.put(FleetGlobal.ACK_IN, ackInFn);
			outputFunctions.put(FleetGlobal.TOKEN_OUT, tokenOutFn);
			outputFunctions.put(FleetGlobal.ACK_TOKEN_IN, ackTokenInFn);
			
			
		} // TokenFifoAction
		
		
		/** Simply pass the data from input terminal 0 */
		class PassData extends OutputFunction {
			
			public PassData() {
				super("PassData");
				Integer fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_FWD_DELAY_DEF;
			}
			
			public Data value() {
				// increase the count of inputs
				counter++;
				// get data from the input
				Object d=  in.getData();
				// is this a payload object?
				if (d instanceof Payload) {
					// get the data item out
					return ((Payload)d).data;
				}
				else if (d instanceof Data) {
					return (Data)d;
				}
				else {
					// wrong input data type, output an illegal value
					director.error("TokenFifoShipElement " + getAlias() + ", id= " + getID() 
							+ "received data of unexpected data type: " 
							+ d + " of type " + d.getClass().getName()
							);						
					return new Data(d, false);
				}
			} // value
		} // PassData

		/** Vacuous function used to acknowldge that the action has fired */ 
		class ActionFired extends OutputFunction {
			
			public ActionFired() {
				super("Vacous");
				Integer fDelay= globals.intValue(TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY);
				functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: TokenFifoShipDefaults.TOKEN_FIFO_SHIP_ACK_DELAY_DEF;
			}
		
			public Data value() {
				return new Data();
			} // value
		} // AcknowledgeData
		
	} // FifoAction	
	
} // TokenFifoShipElement
